import random

import numpy as np
import torch

from outlines_core.kernels.numpy import (
    _apply_token_bitmask_inplace_kernel as numpy_kernel,
)
from outlines_core.kernels.torch import (
    _apply_token_bitmask_inplace_kernel as torch_kernel,
)


def generate_sparse_mask(batch, vocab, allowed_count=1000):
    mask_shape = (batch, (vocab + 31) // 32)
    mask = np.zeros(mask_shape, dtype=np.uint32)
    allowed_indices = random.sample(range(vocab), allowed_count)
    for idx in allowed_indices:
        group = idx // 32
        shift = idx % 32
        bit_mask = np.uint32(1) << np.uint32(shift)
        mask[0, group] |= bit_mask
    return mask


class TorchBitmaskApplyBenchmark:
    params = [[10, 100, 1_000, 10_000, 100_000], [1, 2, 4, 8]]
    param_names = ["allowed_tokens", "batch"]
    number = 10

    def setup(self, allowed_tokens, batch):
        self.device = "cpu"
        self.allowed_tokens = allowed_tokens
        self.vocab = 128000
        self.batch = batch

        self.logits = torch.randn(self.batch, self.vocab, device=self.device)

        mask = torch.from_numpy(
            generate_sparse_mask(
                self.batch, self.vocab, allowed_count=self.allowed_tokens
            )
        )
        self.mask = mask.to(self.device)

        self.kernel = torch_kernel

        for _ in range(4):
            self.kernel(self.logits, self.mask)

    def time_kernel(self, allowed_tokens, batch):
        self.kernel(self.logits, self.mask)


class NumpyBitmaskApplyBenchmark:
    params = [[10, 100, 1_000, 10_000, 100_000], [1, 2, 4, 8]]
    param_names = ["allowed_tokens", "batch"]
    number = 10

    def setup(self, allowed_tokens, batch):
        self.allowed_tokens = allowed_tokens
        self.vocab = 128000
        self.batch = batch

        self.logits = np.random.randn(self.batch, self.vocab).astype(np.float32)

        self.mask = generate_sparse_mask(
            self.batch, self.vocab, allowed_count=self.allowed_tokens
        )

        self.kernel = numpy_kernel

        for _ in range(4):
            self.kernel(self.logits, self.mask)

    def time_kernel(self, allowed_tokens, batch):
        self.kernel(self.logits, self.mask)


class MlxBitmaskApplyBenchmark:
    params = [[10, 100, 1_000, 10_000, 100_000], [1, 2, 4, 8]]
    param_names = ["allowed_tokens", "batch"]
    number = 10

    def setup(self, allowed_tokens, batch):
        try:
            import mlx.core as mx

            from outlines_core.kernels.mlx import (
                _apply_token_bitmask_kernel as mlx_kernel,
            )
        except ImportError:
            raise NotImplementedError

        self.allowed_tokens = allowed_tokens
        self.vocab = 128000
        self.batch = batch

        self.logits = mx.array(
            np.random.randn(self.batch, self.vocab).astype(np.float32)
        )

        self.mask = mx.array(
            generate_sparse_mask(
                self.batch, self.vocab, allowed_count=self.allowed_tokens
            )
        )

        self.kernel = mlx_kernel

        # warm up / compile
        for _ in range(4):
            self.kernel(self.logits, self.mask)

    def time_kernel(self, allowed_tokens, batch):
        self.kernel(self.logits, self.mask)



import copy
import pickle

import pytest

from outlines_core import Vocabulary


@pytest.fixture(scope="session")
def vocabulary():
    eos_token_id = 3
    tokens = {"1": [1], "a": [2]}
    return Vocabulary(eos_token_id, tokens)


def test_basic_vocabulary_interface(vocabulary):
    assert vocabulary.get_eos_token_id() == 3
    assert vocabulary.get("1") == vocabulary.get(b"1") == [1]
    assert len(vocabulary) == 3

    vocabulary.insert("b", 4)
    assert vocabulary.get("b") == [4]
    assert len(vocabulary) == 4

    vocabulary.insert(b"b", 5)
    assert vocabulary.get("b") == vocabulary.get(b"b") == [4, 5]
    assert len(vocabulary) == 5

    vocabulary.remove("b")
    assert vocabulary.get("b") is None

    # second remove doesn't fail too
    vocabulary.remove("b")
    assert vocabulary.get("b") is None

    assert vocabulary.get("a") == [2]
    vocabulary.remove(b"a")
    assert vocabulary.get("a") is None


def test_string_and_bytes_as_tokens():
    eos_token_id = 3
    tokens = {"1": [1], "a": [2]}
    btokens = {b"1": [1], b"a": [2]}
    vocabulary = Vocabulary(eos_token_id, tokens)
    bvocabulary = Vocabulary(eos_token_id, btokens)

    assert (
        vocabulary.get_eos_token_id() == bvocabulary.get_eos_token_id() == eos_token_id
    )
    assert vocabulary.get(b"1") == vocabulary.get("1") == [1]
    assert bvocabulary.get(b"1") == bvocabulary.get("1") == [1]
    assert len(vocabulary) == len(bvocabulary) == 3


def test_do_not_supports_other_types():
    eos_token_id = 0

    with pytest.raises(
        TypeError,
        match=r"Expected a dict with keys of type str or bytes and values of type list\[int\], got",
    ):
        Vocabulary(eos_token_id, 1)

    with pytest.raises(
        TypeError,
        match="Dict keys or/and values of the wrong types",
    ):
        Vocabulary(eos_token_id, {1: [1], 2: [2]})


def test_get_bad_type(vocabulary):
    with pytest.raises(
        TypeError,
        match="Expected a token of type str or bytes, got",
    ):
        vocabulary.get(1)


def test_insert_bad_type(vocabulary):
    with pytest.raises(
        TypeError,
        match="Expected a token of type str or bytes, got",
    ):
        vocabulary.insert(1, 6)


def test_insert_eos_token(vocabulary):
    with pytest.raises(
        ValueError, match="EOS token should not be inserted into Vocabulary"
    ):
        vocabulary.insert("eos-token", 3)


def test_from_pretrained():
    vocabulary = Vocabulary.from_pretrained("gpt2")
    assert vocabulary.get_eos_token_id() == 50256


def test_pickling(vocabulary):
    serialized = pickle.dumps(vocabulary)
    deserialized = pickle.loads(serialized)
    assert deserialized == vocabulary


def test_deepcopy(vocabulary):
    vocabulary2 = copy.deepcopy(vocabulary)
    assert vocabulary2 == vocabulary

    copy_vocabulary2 = copy.deepcopy(vocabulary2)
    assert copy_vocabulary2 == vocabulary2

    vocabulary2.insert("new", 4)
    assert vocabulary2 != copy_vocabulary2
    assert len(vocabulary2) - 1 == len(copy_vocabulary2)
    assert copy_vocabulary2 == vocabulary
    assert len(copy_vocabulary2) == len(vocabulary)



import copy
import pickle
from typing import Dict, List, Union

import pytest

from outlines_core import Guide, Index, Vocabulary


@pytest.fixture(scope="session")
def index() -> Index:
    eos_token_id = 3
    # types here only to please mypy checks
    tokens: Dict[Union[str, bytes], List[int]] = {"1": [1], "2": [2]}
    regex = r"[1-9]"

    vocabulary = Vocabulary(eos_token_id, tokens)
    return Index(regex, vocabulary)


def test_interface():
    eos_token_id = 3
    tokens = {"1": [1], "a": [2]}
    regex = r"[1-9]"

    vocabulary = Vocabulary(eos_token_id, tokens)
    index = Index(regex, vocabulary)
    guide = Guide(index)

    assert guide.get_state() == index.get_initial_state() == 12
    assert guide.get_tokens() == [1]

    assert guide.advance(1) == [vocabulary.get_eos_token_id()]
    assert guide.is_finished()
    assert guide.get_state() == 20
    assert guide.get_tokens() == [eos_token_id]

    with pytest.raises(
        ValueError,
        match="No next state found for the current state",
    ):
        # No advancement is possible for state with allowed tokens == eos
        assert guide.advance(eos_token_id)
        # As well as with any other random token id
        assert guide.advance(4)


def test_regex_final_state_walk():
    # Make sure that the Guide can walk to the final state correctly.
    eos_token_id = 104
    tokens = {b"\n": [103], b".": [102], b"`": [101]}
    regex = r"`\n(\.\n)?`\n"

    vocabulary = Vocabulary(eos_token_id, tokens)
    index = Index(regex, vocabulary)
    guide = Guide(index)

    assert guide.get_tokens() == [101]
    assert guide.advance(101) == [103]
    assert sorted(guide.advance(103)) == [101, 102]
    assert guide.advance(101) == [103]
    assert guide.advance(103) == [vocabulary.get_eos_token_id()]
    assert guide.is_finished()


def test_token_trans_keys_identical():
    tokens = {"a": [1], "b": [2], "z": [3]}
    eos_token_id = 4
    regex = r"z[ab]z"

    vocabulary = Vocabulary(eos_token_id, tokens)
    index = Index(regex, vocabulary)

    guide1 = Guide(index)
    guide2 = Guide(index)

    assert sorted(guide1.advance(3)) == sorted(guide2.advance(3))
    # `a` and `b` have similar transitions to `z`
    assert sorted(guide1.advance(1)) == sorted(guide2.advance(2))
    assert guide1.advance(3) == guide2.advance(3) == [eos_token_id]
    assert guide1.is_finished()
    assert guide2.is_finished()


def test_str_and_bytes_produce_the_same():
    tokens1 = {"a": [1], "b": [2], "z": [3]}
    tokens2 = {b"a": [1], b"b": [2], b"z": [3]}
    eos_token_id = 4
    regex = r"z[ab]z"

    vocabulary1 = Vocabulary(eos_token_id, tokens1)
    vocabulary2 = Vocabulary(eos_token_id, tokens2)
    index1 = Index(regex, vocabulary1)
    index2 = Index(regex, vocabulary2)
    guide1 = Guide(index1)
    guide2 = Guide(index2)

    assert sorted(guide1.advance(3)) == sorted(guide2.advance(3))
    # `a` and `b` have similar transitions to `z`
    assert sorted(guide1.advance(1)) == sorted(guide2.advance(2))
    assert guide1.advance(3) == guide2.advance(3) == [eos_token_id]
    assert guide1.is_finished()
    assert guide2.is_finished()


def test_pickling(index):
    guide = Guide(index)
    serialized = pickle.dumps(guide)
    deserialized = pickle.loads(serialized)
    assert sorted(deserialized.get_tokens()) == sorted(guide.get_tokens())


@pytest.mark.parametrize(
    "model, revision",
    [
        ("openai-community/gpt2", "607a30d783dfa663caf39e06633721c8d4cfcd7e"),
        ("microsoft/phi-2", "ef382358ec9e382308935a992d908de099b64c23"),
        ("Qwen/Qwen1.5-0.5B-Chat", "4d14e384a4b037942bb3f3016665157c8bcb70ea"),
        (
            "NousResearch/Hermes-2-Pro-Llama-3-8B",
            "783fd50eb82d7f57758de033861f54d62dde234f",
        ),
    ],
)
def test_pickling_from_pretrained_with_revision(model, revision):
    regex = "(?:(?:[0-9](?:(?:_)?[0-9])*(?:e|E)(?:(?:\\+|\\-))?[0-9](?:(?:_)?[0-9])*|(?:[0-9](?:(?:_)?[0-9])*\\.(?:[0-9](?:(?:_)?[0-9])*)?|\\.[0-9](?:(?:_)?[0-9])*)(?:(?:e|E)(?:(?:\\+|\\-))?[0-9](?:(?:_)?[0-9])*)?)|[0-9](?:(?:_)?[0-9])*)(?:J|j)|(?:[0-9](?:(?:_)?[0-9])*(?:e|E)(?:(?:\\+|\\-))?[0-9](?:(?:_)?[0-9])*|(?:[0-9](?:(?:_)?[0-9])*\\.(?:[0-9](?:(?:_)?[0-9])*)?|\\.[0-9](?:(?:_)?[0-9])*)(?:(?:e|E)(?:(?:\\+|\\-))?[0-9](?:(?:_)?[0-9])*)?)|0(?:x|X)(?:(?:_)?(?:[0-9]|[a-f]|[A-F]))+|0(?:b|B)(?:(?:_)?[0-1])+|0(?:o|O)(?:(?:_)?[0-7])+|(?:(?i:([ubf]?r?|r[ubf])('([^\\\\']|.)*?'))|(?i:([ubf]?r?|r[ubf])(\"([^\\\"]|.)*?\")))|(?:(?:\r?\n[\t ]*|#[^\n]*))+|[1-9](?:(?:_)?[0-9])*|\\\\[\t \x0c]*\r?\n|continue|nonlocal|assert|global|import|lambda|return|async|await|break|class|False|match|raise|while|yield|case|from|None|pass|True|with|def|del|for|not|try|if|[^\\W\\d]\\w*|#[^\n]*|[\t \x0c]+|\\.\\.\\.|@|\\{|\\(|\\[|\\-|\\+|\\*|\\~"

    vocabulary = Vocabulary.from_pretrained(model, revision=revision)
    index = Index(regex, vocabulary)
    assert len(index.get_transitions()) == 810

    guide = Guide(index)
    serialized = pickle.dumps(guide)
    deserialized = pickle.loads(serialized)
    assert sorted(deserialized.get_tokens()) == sorted(guide.get_tokens())


def test_equality(index):
    guide1 = Guide(index)
    guide2 = Guide(index)
    assert guide1 == guide2

    # confirm that equality is about inner index, not reference difference
    index2 = copy.deepcopy(index)
    guide3 = Guide(index2)
    assert guide3 == guide2 == guide1

    # progress one of the guides, confirm different state == different guide
    guide1.advance(guide1.get_tokens()[-1])
    assert guide1 != guide2
    assert guide3 == guide2


def test_write_mask_into(index):
    import torch

    guide = Guide(index)
    mask = torch.tensor(torch.tensor([-1], dtype=torch.uint32))
    guide.write_mask_into(mask.data_ptr(), mask.numel(), mask.element_size())

    expected_mask = 0
    for token in guide.get_tokens():
        expected_mask |= 1 << (token % 32)

    assert (
        mask[0] == expected_mask
    ), f"mask and expected mask do not match. Mask: {mask[0]}, Expected: {expected_mask}"


def test_write_mask_into_interface(index):
    import torch

    guide = Guide(index)
    mask = torch.tensor(torch.tensor([-1], dtype=torch.uint32))

    with pytest.raises(ValueError, match="Invalid buffer size"):
        guide.write_mask_into(mask.data_ptr(), 0, mask.element_size())
    with pytest.raises(ValueError, match="Invalid element size"):
        guide.write_mask_into(mask.data_ptr(), mask.numel(), 5)
    with pytest.raises(ValueError, match="Invalid data pointer"):
        guide.write_mask_into(0, mask.numel(), mask.element_size())
    with pytest.raises(ValueError, match="Invalid data pointer alignment"):
        guide.write_mask_into(5, mask.numel(), mask.element_size())


def test_rollback(index):
    guide = Guide(index, max_rollback=3)

    first_state = guide.get_state()
    guide.advance(1)

    # Roll back one token to initial state
    guide.rollback_state(1)
    assert not guide.is_finished()
    # we should now be in the initial state
    assert guide.get_state() == first_state


def test_rollback_interface(index):
    guide = Guide(index, max_rollback=3)

    # Rolling back more than recorded history must raise
    with pytest.raises(ValueError, match="Cannot roll back"):
        guide.rollback_state(5)


@pytest.mark.parametrize(
    "seq, expected",
    [
        ([1], True),  # single allowed token: accept
        ([2], True),  # different allowed token: accept
        ([1, 1], False),  # too long for r"[1-9]": reject
        ([2, 3], False),  # extra token: reject
    ],
)
def test_accepts_tokens_correctness(index, seq, expected):
    guide = Guide(index)
    assert guide.accepts_tokens(seq) is expected



from outlines_core import Guide

try:
    import numba
    import numpy as np
except ImportError as e:
    missing_dep = "numba" if "numba" in str(e) else "numpy"
    raise ImportError(
        f"To use the kernels in `outlines_core.kernels.numpy`, `{missing_dep}` must be installed. You can install it with `pip install {missing_dep}`"
    ) from e


def allocate_token_bitmask(vocab_size: int) -> np.ndarray:
    return np.full(
        (1, (vocab_size + 31) // 32),
        -1,
        dtype=np.int32,
    )


@numba.njit
def _apply_token_bitmask_inplace_kernel(logits, mask):
    mask_len = mask.shape[1]
    cutoff = 32 * mask_len

    if logits.shape[1] > cutoff:
        logits[:, cutoff:] = -np.inf
        logits = logits[:, :cutoff]

    n_rows, n_cols = logits.shape

    for i in range(n_rows):
        for mi in range(mask_len):
            mval = mask[i, mi]
            base = mi * 32
            for bit in range(32):
                j = base + bit

                if j >= n_cols:
                    break

                if ((mval >> bit) & 1) == 0:
                    logits[i, j] = -np.inf


def apply_token_bitmask_inplace(logits: np.ndarray, mask: np.ndarray) -> None:
    """
    Apply a logits bitmask inplace, setting the probability of invalid tokens
    to -infinity.

    Arguments:
        logits (np.ndarray): The logits tensor.

        mask (np.ndarray): The token bitmask representing the validity of each
          token in the logits tensor.

    Raises:
        ValueError: If any of the following conditions are not met:
            - `mask.dtype` is not `np.int32`
            - `mask` is not a 2D array
            - `logits` is not a 2D array
            - `mask.shape`shape does not match `logits.shape`

    Returns:
        None: Modifies the mask array in place.
    """
    logits = logits if logits.ndim != 1 else np.expand_dims(logits, axis=0)
    mask = mask if mask.ndim != 1 else np.expand_dims(mask, axis=0)

    if mask.dtype != np.int32:
        raise ValueError(
            f"Invalid mask dtype: Expected `np.int32`, but got `{mask.dtype}`."
        )
    elif mask.ndim != 2:
        raise ValueError(
            f"Invalid mask dimensions: Expected a 2D array, but got {mask.ndim}D."
        )
    elif logits.ndim != 2:
        raise ValueError(
            f"Invalid logits dimensions: Expected a 2D array, but got {logits.ndim}D."
        )
    elif mask.shape[0] != logits.shape[0]:
        raise ValueError(
            f"Invalid batch size: Expected `mask.shape[0]` ({mask.shape[0]}) to match `logits.shape[0]` ({logits.shape[0]})."
        )
    _apply_token_bitmask_inplace_kernel(logits, mask)


def fill_next_token_bitmask(guide: Guide, mask: np.ndarray) -> None:
    """
    Writes a bitmask to represent the tokens permissible by the current state of the `guide`.
    Each bit in the bitmask corresponds to a token ID, with a bit value of 1 indicating that
    the token is allowed and 0 indicating that it is disallowed. This function directly modifies
    the `mask` array in-place.

    Arguments:
        guide (Guide): An instance of the `Guide` class that provides the current guidance state.
        mask (np.ndarray): A 2D tensor of type `torch.int32` where the bitmask will be written.
                             The tensor must be contiguous, have a single batch dimension
                             (shape[0] == 1), and reside on the CPU.

    Raises:
        ValueError: If any of the following conditions are not met:
                    - `mask.dtype` is not `torch.int32`
                    - `mask` is not a 2D tensor
                    - `mask` does not have a single batch dimension (shape[0] != 1)
                    - `mask` is not contiguous in memory
                    - `mask` is not on the CPU device

    Returns:
        None: Modifies the `mask` array in-place.
    """
    if mask.dtype != np.int32:
        raise ValueError(
            f"Invalid mask dtype: Expected `np.int32`, but got `{mask.dtype}`."
        )
    elif mask.ndim != 2:
        raise ValueError(
            f"Invalid mask dimensions: Expected a 2D array, but got {mask.ndim}D."
        )
    elif mask.shape[0] != 1:
        raise ValueError(
            f"Invalid batch size: Batch mask writes are not supported. Expected shape[0] == 1, but got shape {mask.shape}."
        )
    elif not mask.flags["C_CONTIGUOUS"]:
        raise ValueError(
            "Mask array must be contiguous in memory. Use `np.ascontiguousarray(mask)`."
        )

    return guide.write_mask_into(mask.ctypes.data, mask.size, mask.itemsize)



from outlines_core import Guide

try:
    import mlx.core as mx
    import numpy as np
except ImportError as e:
    missing_dep = "numpy" if "numpy" in str(e) else "mlx"
    raise ImportError(
        f"To use the kernels in `outlines_core.kernels.mlx`, {missing_dep} must be installed. You can install it with `pip install {missing_dep}`"
    ) from e


def allocate_token_bitmask(vocab_size: int) -> np.ndarray:
    return np.full(
        (1, (vocab_size + 31) // 32),
        -1,
        dtype=np.int32,
    )


_KERNEL_SOURCE = r"""
// Batch index
uint batch = thread_position_in_grid.y;
// Element index
uint elem = thread_position_in_grid.x;

uint bit = ((elem >> 5) < static_cast<uint>(mask_shape[1])) &&
            ((mask[batch * mask_shape[1] + (elem >> 5)] >> (elem & 31)) & 1);

out[batch * inp_shape[1] + elem] = bit ? inp[batch * inp_shape[1] + elem] : static_cast<T>(-INFINITY);
"""

_KERNEL = mx.fast.metal_kernel(
    name="bitmask_apply_batched",
    input_names=["inp", "mask"],
    output_names=["out"],
    source=_KERNEL_SOURCE,
)


@mx.compile
def _apply_token_bitmask_kernel(data: mx.array, mask: mx.array) -> mx.array:
    return _KERNEL(
        inputs=[data, mask],
        template=[("T", data.dtype)],
        grid=(data.shape[1], data.shape[0], 1),
        threadgroup=(256, 1, 1),
        output_shapes=[data.shape],
        output_dtypes=[data.dtype],
    )[0]


def apply_token_bitmask(logits: mx.array, mask_np: np.ndarray) -> mx.array:
    """
    Apply a logits bitmask inplace, setting the probability of invalid tokens
    to -infinity.

    Arguments:
        logits (mx.array): The logits tensor.

        mask (mx.array): The token bitmask representing the validity of each
          token in the logits tensor.

    Raises:
        ValueError: If any of the following conditions are not met:
            - `mask.dtype` is not `mx.int32`
            - `mask` is not a 2D array
            - `logits` is not a 2D array
            - `mask.shape`shape does not match `logits.shape`

    Returns:
        None: Modifies the mask array in place.
    """
    # makes a copy - non consuming
    mask = mx.array(mask_np)

    logits = logits if len(logits.shape) != 1 else mx.expand_dims(logits, axis=0)
    mask = mask if len(mask.shape) != 1 else mx.expand_dims(mask, axis=0)

    if mask.dtype != mx.int32:
        raise ValueError(
            f"Invalid mask dtype: Expected `np.int32`, but got `{mask.dtype}`."
        )
    elif len(mask.shape) != 2:
        raise ValueError(
            f"Invalid mask dimensions: Expected a 2D array, but got {mask.ndim}D."
        )
    elif len(logits.shape) != 2:
        raise ValueError(
            f"Invalid logits dimensions: Expected a 2D array, but got {logits.ndim}D."
        )
    elif mask.shape[0] != logits.shape[0]:
        raise ValueError(
            f"Invalid batch size: Expected `mask.shape[0]` ({mask.shape[0]}) to match `logits.shape[0]` ({logits.shape[0]})."
        )
    return _apply_token_bitmask_kernel(logits, mask)


def fill_next_token_bitmask(guide: Guide, mask: np.ndarray) -> None:
    """
    Writes a bitmask to represent the tokens permissible by the current state of the `guide`.
    Each bit in the bitmask corresponds to a token ID, with a bit value of 1 indicating that
    the token is allowed and 0 indicating that it is disallowed. This function directly modifies
    the `mask` array in-place.

    Arguments:
        guide (Guide): An instance of the `Guide` class that provides the current guidance state.
        mask (torch.Tensor): A 2D tensor of type `torch.int32` where the bitmask will be written.
                             The tensor must be contiguous, have a single batch dimension
                             (shape[0] == 1), and reside on the CPU.

    Raises:
        ValueError: If any of the following conditions are not met:
                    - `mask.dtype` is not `np.int32`
                    - `mask` is not a 2D tensor
                    - `mask` does not have a single batch dimension (shape[0] != 1)
                    - `mask` is not contiguous in memory
                    - `mask` is not on the CPU device

    Returns:
        None: Modifies the `mask` tensor in-place.
    """
    if mask.dtype != np.int32:
        raise ValueError(
            f"Invalid mask dtype: Expected `np.int32`, but got `{mask.dtype}`."
        )
    elif mask.ndim != 2:
        raise ValueError(
            f"Invalid mask dimensions: Expected a 2D array, but got {mask.ndim}D."
        )
    elif mask.shape[0] != 1:
        raise ValueError(
            f"Invalid batch size: Batch mask writes are not supported. Expected shape[0] == 1, but got shape {mask.shape}."
        )
    elif not mask.flags["C_CONTIGUOUS"]:
        raise ValueError(
            "Mask array must be contiguous in memory. Use `np.ascontiguousarray(mask)`."
        )

    return guide.write_mask_into(mask.ctypes.data, mask.size, mask.itemsize)



import sys

# There is a conflict between asv.statistics and the standard library's statistics module.
# This is a workaround to use the standard library's median function.
if "asv.statistics" in sys.modules:

    def median(data):
        import statistics

        return statistics.median(data)

    asv_statistics = sys.modules["asv.statistics"]
    asv_statistics.median = median  # type: ignore



import copy
import gc
import pickle

import pytest

from outlines_core import Index, Vocabulary


@pytest.fixture(scope="session")
def index() -> Index:
    eos_token_id = 3
    tokens = {"1": [1], "2": [2]}
    regex = r"[1-9]"

    vocabulary = Vocabulary(eos_token_id, tokens)
    return Index(regex, vocabulary)


def test_basic_interface(index):
    init_state = index.get_initial_state()
    assert init_state == 12
    assert index.is_final_state(init_state) is False

    allowed_tokens = index.get_allowed_tokens(init_state)
    assert allowed_tokens == [1, 2]

    next_state = index.get_next_state(init_state, allowed_tokens[-1])
    assert next_state == 20
    assert index.is_final_state(next_state) is True
    assert index.get_final_states() == {20}

    expected_transitions = {
        12: {
            1: 20,
            2: 20,
        },
        20: {
            3: 20,
        },
    }
    assert index.get_transitions() == expected_transitions


def test_pickling(index):
    serialized = pickle.dumps(index)
    deserialized = pickle.loads(serialized)
    assert deserialized == index


def test_deepcopy(index):
    index2 = copy.deepcopy(index)
    assert index2 == index

    copy_index2 = copy.deepcopy(index2)
    assert copy_index2 == index2

    index2_id = id(index2)
    del index2
    gc.collect()
    is_deleted = not any(id(o) == index2_id for o in gc.get_objects())
    assert is_deleted

    assert copy_index2 == index



import os
from concurrent.futures import ThreadPoolExecutor

import psutil

from outlines_core import Guide, Index, Vocabulary

regex_samples = {
    "email": r"[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?",
    "complex_phone": "\\+?\\d{1,4}?[-.\\s]?\\(?\\d{1,3}?\\)?[-.\\s]?\\d{1,4}[-.\\s]?\\d{1,4}[-.\\s]?\\d{1,9}",
    "simple_phone": "\\+?[1-9][0-9]{7,14}",
    "date": r"([1-9]|0[1-9]|1[0-9]|2[0-9]|3[0-1])(\.|-|/)([1-9]|0[1-9]|1[0-2])(\.|-|/)([0-9][0-9]|19[0-9][0-9]|20[0-9][0-9])|([0-9][0-9]|19[0-9][0-9]|20[0-9][0-9])(\.|-|/)([1-9]|0[1-9]|1[0-2])(\.|-|/)([1-9]|0[1-9]|1[0-9]|2[0-9]|3[0-1])",
    "time": r"(0?[1-9]|1[0-2]):[0-5]\d\s?(am|pm)?",
    "ip": r"(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)",
    "url": r"(https?:\/\/)?([\da-z\.-]+)\.([a-z\.]{2,6})([\/\w \.-]*)*\/?",
    "ssn": r"\d{3}-\d{2}-\d{4}",
    "complex_span_constrained_relation_extraction": "(['\"\\ ,]?((?:of|resulting|case|which|cultures|a|core|extreme|selflessness|spiritual|various|However|both|vary|in|other|secular|the|religious|among|moral|and|It|object|worldviews|altruism|traditional|material|aspect|or|life|beings|virtue|is|however|opposite|concern|an|practice|it|for|s|quality|religions|In|Altruism|animals|happiness|many|become|principle|human|selfishness|may|synonym)['\"\\ ,]?)+['\"\\ ,]?\\s\\|\\s([^|\\(\\)\n]{1,})\\s\\|\\s['\"\\ ,]?((?:of|resulting|case|which|cultures|a|core|extreme|selflessness|spiritual|various|However|both|vary|in|other|secular|the|religious|among|moral|and|It|object|worldviews|altruism|traditional|material|aspect|or|life|beings|virtue|is|however|opposite|concern|an|practice|it|for|s|quality|religions|In|Altruism|animals|happiness|many|become|principle|human|selfishness|may|synonym)['\"\\ ,]?)+['\"\\ ,]?(\\s\\|\\s\\(([^|\\(\\)\n]{1,})\\s\\|\\s([^|\\(\\)\n]{1,})\\))*\\n)*",
}


class RegexIndexBenchmark:
    params = regex_samples.keys()

    def setup(self, pattern_name):
        self.vocabulary = Vocabulary.from_pretrained("gpt2")
        self.pattern = regex_samples[pattern_name]

    def time_regex_to_guide(self, pattern_name):
        Index(self.pattern, self.vocabulary)

    def time_regex_to_guide_threads(self, pattern_name):
        # Default GIL switch interval is 5ms (0.005), which isn't helpful for cpu heavy tasks,
        # this parallel case should be relatively close in runtime to one thread, but it is not,
        # because of the GIL.
        core_count = psutil.cpu_count(logical=False)
        with ThreadPoolExecutor(max_workers=core_count) as executor:
            list(executor.map(self._from_regex, [pattern_name] * core_count))

    def time_regex_to_guide_threads_with_custom_switch_interval(self, pattern_name):
        # Note: after moving to full rust implementation for index and guide creation, this experiment
        # is no longer shows the drastic difference as it once showed when python was heavily involved,
        # due to average speedup ~10 times.

        # This test is to show, that if GIL's switch interval is set to be longer, then the parallel
        # test's runtime on physical cores will be much closer to the one-threaded case.
        import sys

        sys.setswitchinterval(5)

        core_count = psutil.cpu_count(logical=False)
        with ThreadPoolExecutor(max_workers=core_count) as executor:
            list(executor.map(self._from_regex, [pattern_name] * core_count))

    def _from_regex(self, pattern_name):
        Index(self.pattern, self.vocabulary)


class MemoryRegexIndexBenchmark:
    params = ["simple_phone", "complex_span_constrained_relation_extraction"]

    def setup(self, pattern_name):
        self.vocabulary = Vocabulary.from_pretrained("gpt2")
        self.pattern = regex_samples[pattern_name]

    def peakmem_regex_to_index(self, pattern_name):
        Index(self.pattern, self.vocabulary)


class MemoryStabilityBenchmark:
    params = [1, 10_000]

    def setup(self, num):
        self.vocabulary = Vocabulary.from_pretrained("gpt2")
        self.index = Index(".*", self.vocabulary)
        self.process = psutil.Process(os.getpid())

    def _memory_usage(self):
        return self.process.memory_info().rss / 1024**2

    def peakmem_guides_per_index(self, num_guides):
        initial = self._memory_usage()
        objects = [Guide(self.index) for i in range(num_guides)]
        final = self._memory_usage()

        assert len(objects) == num_guides
        assert final - initial < 5


class WriteMaskIntoBenchmark:
    params = list(regex_samples.keys())
    param_names = ["regex_key"]

    def setup(self, regex_key):
        from outlines_core.kernels.torch import allocate_token_bitmask

        self.vocab = Vocabulary.from_pretrained("gpt2")
        self.mask = allocate_token_bitmask(len(self.vocab))
        self.index = Index(regex_samples[regex_key], self.vocab)
        self.guide = Guide(self.index)

    def time_write_mask_into(self, regex_key):
        self.guide.write_mask_into(
            self.mask.data_ptr(), self.mask.numel(), self.mask.element_size()
        )



"""Tests for package imports to catch import/module registration issues."""


def test_import_outlines_core():
    import outlines_core

    assert hasattr(outlines_core, "Guide")
    assert hasattr(outlines_core, "Index")
    assert hasattr(outlines_core, "Vocabulary")
    assert hasattr(outlines_core, "json_schema")


def test_import_json_schema_module():
    from outlines_core import json_schema

    assert hasattr(json_schema, "BOOLEAN")
    assert hasattr(json_schema, "build_regex_from_schema")


def test_import_from_json_schema():
    from outlines_core.json_schema import (  # noqa: F401
        BOOLEAN,
        DATE,
        DATE_TIME,
        EMAIL,
        INTEGER,
        NULL,
        NUMBER,
        STRING,
        STRING_INNER,
        TIME,
        URI,
        UUID,
        WHITESPACE,
        build_regex_from_schema,
    )

    assert BOOLEAN == "(true|false)"
    assert callable(build_regex_from_schema)


def test_import_main_classes():
    from outlines_core import Guide, Index, Vocabulary

    assert Guide is not None
    assert Index is not None
    assert Vocabulary is not None


# The tests below ensure that users can import using
# `outlines_core.outlines_core`. This is not the intended usage but we should
# keep supporting it as users may have fallen back to it when the regular
# Python imports were buggy (fixed in v0.2.14)


def test_import_rust_extension_directly():
    from outlines_core import outlines_core

    assert hasattr(outlines_core, "Guide")
    assert hasattr(outlines_core, "Index")
    assert hasattr(outlines_core, "Vocabulary")
    assert hasattr(outlines_core, "json_schema")


def test_import_from_rust_extension():
    from outlines_core.outlines_core import Guide, Index, Vocabulary

    assert Guide is not None
    assert Index is not None
    assert Vocabulary is not None


def test_import_json_schema_from_rust_extension():
    from outlines_core.outlines_core import json_schema

    assert hasattr(json_schema, "BOOLEAN")
    assert hasattr(json_schema, "build_regex_from_schema")



import sys

# kernels is not reexported as it should remain an optional dependency
from . import _json_schema as json_schema
from .outlines_core import Guide, Index, Vocabulary

# Register json_schema in sys.modules so "from outlines_core.json_schema
# import ..." works
sys.modules["outlines_core.json_schema"] = json_schema



# Re-export from Rust extension
from .outlines_core import json_schema as _json_schema

BOOLEAN = _json_schema.BOOLEAN
DATE = _json_schema.DATE
DATE_TIME = _json_schema.DATE_TIME
EMAIL = _json_schema.EMAIL
INTEGER = _json_schema.INTEGER
NULL = _json_schema.NULL
NUMBER = _json_schema.NUMBER
STRING = _json_schema.STRING
STRING_INNER = _json_schema.STRING_INNER
TIME = _json_schema.TIME
URI = _json_schema.URI
UUID = _json_schema.UUID
WHITESPACE = _json_schema.WHITESPACE
build_regex_from_schema = _json_schema.build_regex_from_schema

__all__ = [
    "BOOLEAN",
    "DATE",
    "DATE_TIME",
    "EMAIL",
    "INTEGER",
    "NULL",
    "NUMBER",
    "STRING",
    "STRING_INNER",
    "TIME",
    "URI",
    "UUID",
    "WHITESPACE",
    "build_regex_from_schema",
]



import json
import re

import pytest
from pydantic import BaseModel

from outlines_core.json_schema import build_regex_from_schema


def test_build_regex_from_json_schema():
    class FooBar(BaseModel):
        foo: int
        bar: str

    schema = json.dumps(FooBar.model_json_schema())

    regex = build_regex_from_schema(schema)
    expected = """{"foo" : 4 ,"bar":"baz    baz baz bar"}"""
    assert re.fullmatch(regex, expected)

    # any whitespace pattern can be used
    regex = build_regex_from_schema(schema, r"[\n ]*")
    expected = """{     "foo"   :   4, \n\n\n   "bar": "baz    baz baz bar"\n\n}"""
    assert re.fullmatch(regex, expected)


def test_invalid_json():
    with pytest.raises(
        TypeError,
        match="Expected a valid JSON string.",
    ):
        build_regex_from_schema("{'name':")


def test_types_presence_and_not_emptyness():
    from outlines_core.json_schema import (
        BOOLEAN,
        DATE,
        DATE_TIME,
        EMAIL,
        INTEGER,
        NULL,
        NUMBER,
        STRING,
        STRING_INNER,
        TIME,
        URI,
        UUID,
        WHITESPACE,
    )

    assert BOOLEAN
    assert DATE
    assert DATE_TIME
    assert EMAIL
    assert INTEGER
    assert NULL
    assert NUMBER
    assert STRING
    assert STRING_INNER
    assert TIME
    assert URI
    assert UUID
    assert WHITESPACE



import importlib

import numpy as np
import pytest
import torch

from outlines_core import Guide, Index, Vocabulary

VOCAB = Vocabulary.from_pretrained("gpt2", None, None)
VOCAB_LEN = len(VOCAB)


@pytest.fixture(scope="session")
def guide() -> Guide:
    return Guide(Index("\\+?[1-9][0-9]{7,14}", VOCAB))


@pytest.mark.no_cover
def test_interface_torch():
    from outlines_core.kernels.torch import (
        allocate_token_bitmask,
        apply_token_bitmask_inplace,
        fill_next_token_bitmask,
    )

    mask = allocate_token_bitmask(VOCAB_LEN)
    assert mask.shape[1] == 1571, "Mask sized incorrectly."
    assert mask.dim() == 2, "Mask should be 2D"
    assert mask.is_contiguous(), "Mask must be contiguous"
    assert mask.dtype == torch.int32, "Mask must be dtype torch.int32"

    # test apply_token_bitmask_inplace
    logits = torch.randn(1, 100)

    wrong_dtype_mask = mask.to(torch.int64)
    with pytest.raises(ValueError, match="Invalid mask dtype"):
        apply_token_bitmask_inplace(logits, wrong_dtype_mask)

    mask_1d = torch.full((mask.shape[1],), -1, dtype=torch.int32)
    with pytest.raises(
        ValueError, match="Invalid mask dimensions: Expected a 2D array"
    ):
        apply_token_bitmask_inplace(logits, mask_1d)

    logits_1d = torch.randn(100)
    with pytest.raises(
        ValueError, match="Invalid logits dimensions: Expected a 2D array, but got 1D."
    ):
        apply_token_bitmask_inplace(logits_1d, mask)

    mask_batch2 = torch.full((2, mask.shape[1]), -1, dtype=torch.int32)
    with pytest.raises(ValueError, match="Invalid batch size"):
        apply_token_bitmask_inplace(logits, mask_batch2)

    # test fill_next_token_bitmask
    wrong_dtype_mask = mask.to(torch.int64)
    with pytest.raises(ValueError, match="Invalid mask dtype"):
        fill_next_token_bitmask(None, wrong_dtype_mask)

    mask_1d = torch.full((mask.shape[1],), -1, dtype=torch.int32)
    with pytest.raises(ValueError, match="Invalid mask dimensions"):
        fill_next_token_bitmask(None, mask_1d)

    mask_batch2 = torch.full((2, mask.shape[1]), -1, dtype=torch.int32)
    with pytest.raises(
        ValueError, match="Invalid batch size: Batch mask writes are not supported"
    ):
        fill_next_token_bitmask(None, mask_batch2)


@pytest.mark.no_cover
def test_interface_numpy():
    from outlines_core.kernels.numpy import (
        allocate_token_bitmask,
        apply_token_bitmask_inplace,
        fill_next_token_bitmask,
    )

    mask = allocate_token_bitmask(VOCAB_LEN)
    assert mask.shape[1] == 1571, "Mask sized incorrectly."
    assert mask.ndim == 2, "Mask should be 2D"
    assert mask.flags["C_CONTIGUOUS"], "Mask must be contiguous"
    assert mask.dtype == np.int32, "Mask must be dtype np.int32"

    logits = np.random.randn(1, 100).astype(np.float32)

    wrong_dtype_mask = mask.astype(np.int64)
    with pytest.raises(ValueError, match="Invalid mask dtype"):
        apply_token_bitmask_inplace(logits, wrong_dtype_mask)

    # 3d because 1d will be fixed automatically
    mask_3d = np.full((mask.shape[1], 1, mask.shape[1]), -1, dtype=np.int32)
    with pytest.raises(
        ValueError, match="Invalid mask dimensions: Expected a 2D array"
    ):
        apply_token_bitmask_inplace(logits, mask_3d)

    with pytest.raises(
        ValueError, match="Invalid logits dimensions: Expected a 2D array"
    ):
        apply_token_bitmask_inplace(mask_3d, mask)

    mask_batch2 = np.full((2, mask.shape[1]), -1, dtype=np.int32)
    with pytest.raises(ValueError, match="Invalid batch size"):
        apply_token_bitmask_inplace(logits, mask_batch2)

    wrong_dtype_mask = mask.astype(np.int64)
    with pytest.raises(ValueError, match="Invalid mask dtype"):
        fill_next_token_bitmask(None, wrong_dtype_mask)

    mask_1d = np.full((mask.shape[1],), -1, dtype=np.int32)
    with pytest.raises(ValueError, match="Invalid mask dimensions"):
        fill_next_token_bitmask(None, mask_1d)

    mask_batch2 = np.full((2, mask.shape[1]), -1, dtype=np.int32)
    with pytest.raises(
        ValueError, match="Invalid batch size: Batch mask writes are not supported"
    ):
        fill_next_token_bitmask(None, mask_batch2)


@pytest.mark.no_cover
@pytest.mark.skipif(
    not importlib.util.find_spec("mlx"), reason="mlx is required to test mlx kernels"  # type: ignore
)
def test_interface_mlx():
    import mlx.core as mx
    import numpy as np
    import pytest

    from outlines_core.kernels.mlx import (
        allocate_token_bitmask,
        apply_token_bitmask,
        fill_next_token_bitmask,
    )

    mask = allocate_token_bitmask(VOCAB_LEN)
    assert mask.shape[1] == (VOCAB_LEN + 31) // 32, "Mask sized incorrectly."
    assert mask.ndim == 2, "Mask should be 2D"
    assert mask.flags["C_CONTIGUOUS"], "Mask must be contiguous"
    assert mask.dtype == np.int32, "Mask must be dtype np.int32"

    logits = mx.array(np.random.randn(1, 100).astype(np.float32))
    wrong_dtype_mask = mask.astype(np.int64)
    with pytest.raises(ValueError, match="Invalid mask dtype"):
        apply_token_bitmask(logits, wrong_dtype_mask)

    # 3d because 1d will be fixed automatically
    mask_3d = np.full((mask.shape[1], 1, mask.shape[1]), -1, dtype=np.int32)
    with pytest.raises(
        ValueError, match="Invalid mask dimensions: Expected a 2D array"
    ):
        apply_token_bitmask(logits, mask_3d)

    with pytest.raises(
        ValueError, match="Invalid logits dimensions: Expected a 2D array"
    ):
        apply_token_bitmask(mask_3d, mask)

    mask_batch2 = np.full((2, mask.shape[1]), -1, dtype=np.int32)
    with pytest.raises(ValueError, match="Invalid batch size"):
        apply_token_bitmask(logits, mask_batch2)

    wrong_dtype_mask = mask.astype(np.int64)
    with pytest.raises(ValueError, match="Invalid mask dtype"):
        fill_next_token_bitmask(None, wrong_dtype_mask)

    mask_1d = np.full((mask.shape[1],), -1, dtype=np.int32)
    with pytest.raises(ValueError, match="Invalid mask dimensions"):
        fill_next_token_bitmask(None, mask_1d)

    mask_batch2 = np.full((2, mask.shape[1]), -1, dtype=np.int32)
    with pytest.raises(
        ValueError, match="Invalid batch size: Batch mask writes are not supported"
    ):
        fill_next_token_bitmask(None, mask_batch2)


@pytest.mark.no_cover
def test_torch_correctness(guide):
    from outlines_core.kernels.torch import _apply_token_bitmask_inplace_kernel

    allowed_tokens = set(guide.get_tokens())

    logits = torch.tensor(torch.randn(1, VOCAB_LEN))

    orig_logits = logits.clone()

    mask = torch.tensor(
        torch.full((1, ((VOCAB_LEN + 31) // 32)), -1, dtype=torch.int32)
    )

    guide.write_mask_into(mask.data_ptr(), mask.numel(), mask.element_size())

    _apply_token_bitmask_inplace_kernel(logits, mask)

    for j in range(VOCAB_LEN):
        if j in allowed_tokens:
            assert torch.isclose(
                logits[0, j], orig_logits[0, j], equal_nan=True
            ), f"Token {j} should be allowed but was masked."
        else:
            assert logits[0, j] == -float(
                "inf"
            ), f"Token {j} should be masked but was {logits[0, j].item()}."


@pytest.mark.no_cover
def test_numpy_correctness(guide):
    from outlines_core.kernels.numpy import _apply_token_bitmask_inplace_kernel

    allowed_tokens = set(guide.get_tokens())

    logits = np.random.randn(1, VOCAB_LEN).astype(np.float32)
    orig_logits = logits.copy()

    mask = np.full((1, ((VOCAB_LEN + 31) // 32)), -1, dtype=np.int32)

    guide.write_mask_into(mask.ctypes.data, mask.size, mask.itemsize)

    _apply_token_bitmask_inplace_kernel(logits, mask)

    for j in range(VOCAB_LEN):
        if j in allowed_tokens:
            np.testing.assert_allclose(
                logits[0, j],
                orig_logits[0, j],
                err_msg=f"Token {j} should be allowed but was masked.",
            )
        else:
            assert (
                logits[0, j] == -np.inf
            ), f"Token {j} should be masked was got {logits[0, j]}."


@pytest.mark.no_cover
@pytest.mark.skipif(
    not importlib.util.find_spec("mlx"), reason="mlx is required to test mlx kernels"  # type: ignore
)
def test_mlx_correctness(guide):
    import mlx.core as mx

    from outlines_core.kernels.mlx import _apply_token_bitmask_kernel

    allowed_tokens = set(guide.get_tokens())

    np_logits = np.random.randn(1, VOCAB_LEN).astype(np.float32)

    orig_logits = np_logits.copy()

    logits_mlx = mx.array(np_logits)

    mask = np.full((1, (VOCAB_LEN + 31) // 32), -1, dtype=np.int32)

    guide.write_mask_into(mask.ctypes.data, mask.size, mask.itemsize)

    logits_mlx_out = _apply_token_bitmask_kernel(logits_mlx, mx.array(mask))

    logits_out = np.array(logits_mlx_out)

    for j in range(VOCAB_LEN):
        if j in allowed_tokens:
            np.testing.assert_allclose(
                logits_out[0, j],
                orig_logits[0, j],
                err_msg=f"Token {j} should be allowed but was masked.",
            )
        else:
            assert (
                logits_out[0, j] == -np.inf
            ), f"Token {j} should be masked was got {logits_out[0, j]}."



# Provides kernels for masking a logits tensor,
# using the write_into_mask method on the `Guide` object and the bitmask
# which it writes into a tensor.
#
# Kernels inspired by https://github.com/guidance-ai/llguidance/blob/main/python/llguidance/torch.py
from outlines_core import Guide

try:
    import torch
except Exception as e:
    raise ImportError(
        "`torch` is required to use the kernels from"
        "`outlines_core.kernels.torch. You can install "
        "`torch` using the official guide at https://pytorch.org/get-started/locally/"
    ) from e


def allocate_token_bitmask(vocab_size: int) -> torch.Tensor:
    """
    Allocate a token bitmask for use with the `Guide.write_into_mask` API and logits masking,
    based on the vocab_size.

    Arguments:
        - vocab_size: int
    Returns:
        -  torch.Tensor
    """
    return torch.full(
        (1, (vocab_size + 31) // 32),
        -1,
        dtype=torch.int32,
        pin_memory=torch.cuda.is_available(),
    )


# This takes roughly 23 microseconds per run, with a bitmask of
# 1k allowed tokens, and 128k logits tensor.
# Also compiles to one graph with no graph breaks
# Performance characteristics are:
# - Larger the logits array ( length ), the longer the kernel takes
# - Constant time for mask i.e. number of allowed tokens does not affect execution
#   time
@torch.compile(dynamic=True)
def _apply_token_bitmask_inplace_kernel(logits, mask):
    # This will set any logits beyond the mask
    # to -torch.inf
    cutoff = 32 * mask.shape[1]
    logits[:, cutoff:] = -torch.inf

    # Unpack mask so each bit is a boolean
    bit_masks = (
        (
            torch.bitwise_right_shift(
                mask.unsqueeze(-1),
                torch.arange(32, device=mask.device, dtype=torch.int32),
            )
            & 1
        )
        .bool()
        .view(mask.shape[0], -1)
        .narrow(1, 0, logits.shape[1])
    )

    logits.masked_fill_(~bit_masks, -torch.inf)


def apply_token_bitmask_inplace(logits: torch.Tensor, mask: torch.Tensor) -> None:
    """
    Apply a logits bitmask inplace, setting the probability of invalid tokens
    to -infinity.

    Arguments:
        logits (torch.Tensor): The logits tensor.

        mask (torch.Tensor): The token bitmask representing the validity of
          each token in the logits tensor.

    Raises:
        ValueError: If any of the following conditions are not met:
            - `mask.dtype` is not `torch.int32`
            - `mask` is not a 2D array
            - `logits` is not a 2D array
            - `mask.shape`shape does not match `logits.shape`

    Returns:
        None: Modifies the logits tensor in place
    """
    if mask.dtype != torch.int32:
        raise ValueError(
            f"Invalid mask dtype: Expected `torch.int32`, but got `{mask.dtype}`."
        )
    elif mask.dim() != 2:
        raise ValueError(
            f"Invalid mask dimensions: Expected a 2D array, but got {mask.dim()}D."
        )
    elif logits.dim() != 2:
        raise ValueError(
            f"Invalid logits dimensions: Expected a 2D array, but got {logits.dim()}D."
        )
    elif mask.shape[0] != logits.shape[0]:
        raise ValueError(
            f"Invalid batch size: Expected `mask.shape[0]` ({mask.shape[0]}) to match `logits.shape[0]` ({logits.shape[0]})."
        )
    _apply_token_bitmask_inplace_kernel(logits, mask)


def fill_next_token_bitmask(guide: Guide, mask: torch.Tensor) -> None:
    """
    Writes a bitmask to represent the tokens permissible by the current state of the `guide`.
    Each bit in the bitmask corresponds to a token ID, with a bit value of 1 indicating that
    the token is allowed and 0 indicating that it is disallowed. This function directly modifies
    the `mask` tensor in-place.

    Arguments:
        guide (Guide): An instance of the `Guide` class that provides the current guidance state.
        mask (torch.Tensor): A 2D tensor of type `torch.int32` where the bitmask will be written.
                             The tensor must be contiguous, have a single batch dimension
                             (shape[0] == 1), and reside on the CPU.

    Raises:
        ValueError: If any of the following conditions are not met:
                    - `mask.dtype` is not `torch.int32`
                    - `mask` is not a 2D tensor
                    - `mask` does not have a single batch dimension (shape[0] != 1)
                    - `mask` is not contiguous in memory
                    - `mask` is not on the CPU device

    Returns:
        None: Modifies the `mask` tensor in-place.
    """
    if mask.dtype != torch.int32:
        raise ValueError(
            f"Invalid mask dtype: Expected `torch.int32`, but got `{mask.dtype}`."
        )
    elif mask.dim() != 2:
        raise ValueError(
            f"Invalid mask dimensions: Expected a 2D array, but got {mask.dim()}D."
        )
    elif mask.shape[0] != 1:
        raise ValueError(
            f"Invalid batch size: Batch mask writes are not supported. Expected shape[0] == 1, but got shape {mask.shape}."
        )
    elif not mask.is_contiguous():
        raise ValueError(
            "Mask array must be contiguous in memory. Use `mask.contiguous()` to fix it."
        )
    elif mask.device != torch.device("cpu"):
        raise ValueError(
            f"Invalid device: Expected `mask` tensor to be on device `cpu`, but found it on `{mask.device}`."
        )

    guide.write_mask_into(mask.data_ptr(), mask.numel(), mask.element_size())
